/**************************************************************************
**       Title: simple RGB image
**    $RCSfile: SimpleRGBImage.cc,v $
**   $Revision: 1.7 $$Name:  $
**       $Date: 2002/11/16 09:27:57 $
**   Copyright: GPL $Author: pigorsch $
** Description:
**
**    
**
**-------------------------------------------------------------------------
**
**  $Log: SimpleRGBImage.cc,v $
**  Revision 1.7  2002/11/16 09:27:57  pigorsch
**  - fixed read function
**
**  Revision 1.6  2002/11/13 11:26:06  pigorsch
**  *** empty log message ***
**
**  Revision 1.5  2002/11/05 16:06:17  pigorsch
**  - corrected typo
**
**  Revision 1.4  2002/11/05 15:19:20  pigorsch
**  - check if the image itself is assigned by operator=
**
**  Revision 1.3  2002/11/05 15:02:32  pigorsch
**  - removed size checks before delete[]
**
**  Revision 1.2  2002/11/05 14:44:19  pigorsch
**  - using operator= in copy constructor
**
**  Revision 1.1  2002/11/02 12:07:11  pigorsch
**  - initial revision
**
**
**
**************************************************************************/

#include "SimpleRGBImage.hh"

#include <iomanip>
#include <iostream>
#include <string>
#include <cctype>

/*======================================================================*/
/*! 
 *   eats white spaces and comments (starting with '#'
 *   until line end) from the given stream
 *
 *   \param  s  stream
 *
 *======================================================================*/
std::istream& eatWsAndComments(std::istream &s)
{
  s >> std::ws;
  
  while (s.peek()=='#')
  { 
    s.ignore(10000, '\n');  // ignore characters until line end
    s >> std::ws;
  }
  return s;
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  Default Constructor
 *  ==> see headerfile
 *=======================================================================*/
SimpleRGBImage::
SimpleRGBImage()
        :_width(0), _height(0), _size(0),
         _pixels(0), _pixelsEnd(0),
         _rowStart(0)
{}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  Copy Constructor
 *  ==> see headerfile
 *=======================================================================*/
SimpleRGBImage::
SimpleRGBImage(const SimpleRGBImage& image)
        :_width(0), _height(0), _size(0),
         _pixels(0), _pixelsEnd(0),
         _rowStart(0)
{
  (*this)=image;
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  Size Constructor
 *  ==> see headerfile
 *=======================================================================*/
SimpleRGBImage::
SimpleRGBImage(unsigned int width, unsigned int height)
        :_width(0), _height(0), _size(0),
         _pixels(0), _pixelsEnd(0),
         _rowStart(0)
{
  resize(width, height);
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  Load Constructor
 *  ==> see headerfile
 *=======================================================================*/
SimpleRGBImage::
SimpleRGBImage(std::istream& is)
        :_width(0), _height(0), _size(0),
         _pixels(0), _pixelsEnd(0),
         _rowStart(0)
{
  read(is);
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  Destructor
 *  ==> see headerfile
 *=======================================================================*/
SimpleRGBImage::
~SimpleRGBImage()
{
  delete[] _pixels;
  delete[] _rowStart;
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  resize(...)
 *  ==> see headerfile
 *=======================================================================*/
void 
SimpleRGBImage::
resize(unsigned int width, unsigned int height)
{
  if (width>MAX_DIMENSION || height>MAX_DIMENSION)
  {
    MyError e("Desired dimensions are too large");
    throw e;
  }

  if (width==0 || height==0)
  {
    width=0;
    height=0;
  }
  
  if (width!=_width || height!=_height)
  {
    unsigned int newSize=width*height;
    _width=width;
    
    if (_size!=newSize)
    {
      delete[] _pixels;
      _pixels=new RGBColor[newSize];
      if (!_pixels)
      {
        MyError e("Not enough memory available to store image");
        throw e;
      }
      
      _pixelsEnd=_pixels+newSize;
      _size=newSize;
    }

    if (_height!=height)
    {
      delete[] _rowStart;
      _rowStart=new RGBColor*[height];
      if (!_rowStart)
      {
        MyError e("Not enough memory available to store image");
        throw e;
      }

      _height=height;
    }

    RGBColor* itPixel=_pixels;
    for (unsigned int i=0; i<_height; ++i)
    {
      _rowStart[i]=itPixel;
      itPixel+=_width;
    }   
  }
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  read(...)
 *  ==> see headerfile
 *=======================================================================*/
void 
SimpleRGBImage::
read(std::istream& is)
{
  std::string header;
  is >> std::setw(2+1) >> header;
  if (!is.good())
  {
    ReadError e("Error while reading P3");
    throw e;
  }
  
  if (header!="P3")
  {
    ReadError e("Unknown header " + header + ", expected P3");
    throw e;
  }
  
  /*-------------------------------------------------------------------
   *  read width, height and maximum color-component value
   *-------------------------------------------------------------------*/
  int width;
  is >> eatWsAndComments >> width;
  if (!is.good()) 
  {
    throw ReadError("error while reading image width");
  }
  
  int height;
  is >> eatWsAndComments >> height;
  if (!is.good()) 
  {
    throw ReadError("error while reading image height\n");
  }
  
  int maxVal;
  is >> eatWsAndComments >> maxVal;
  if (!is.good() || maxVal>255)
  {
    throw ReadError("only maxVal<=255 supported");
  }
  
  /*-------------------------------------------------------------------
   *  resize memory for image if size changed
   *-------------------------------------------------------------------*/
  resize(width, height);
  
  for (RGBColor* p=begin();
       p!=end();
       ++p)
  {
    if (!is.good()) 
    {
      throw ReadError("error while reading image pixels");
    }

    unsigned int r, g, b;
    is >> r >> g >> b;

    if (r>255 || g>255 || b>255)
    {
      throw ReadError("color channel value to large");
    }

    p->set(r, g, b);  
  }
  
  //is >> ws;  // eat trailing whitespaces, if there are any
  is >> eatWsAndComments;
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  write(...)
 *  ==> see headerfile
 *=======================================================================*/
void 
SimpleRGBImage::
write(std::ostream& os)
{
  os << "P3\n"
     << width() << " " << height() << "\n"
     << "255\n";
  
  for (const RGBColor* p=begin();
       p!=end();
       ++p)
  {
    os << (unsigned int)p->red() << "\n" 
       << (unsigned int)p->green() << "\n" 
       << (unsigned int)p->blue() << "\n" ;
  }
}

/*=========================================================================
 *  DESCRIPTION OF FUNCTION:  operator=(...)
 *  ==> see headerfile
 *=======================================================================*/
SimpleRGBImage& 
SimpleRGBImage::
operator=(const SimpleRGBImage& image)
{
  if (this != (&image))
  {
    resize(image.width(), image.height());
    
    const RGBColor* p1=image.begin();
    RGBColor* p2=begin();
    
    while (p1!=image.end())
    {
      *p2=*p1;
      ++p1;
      ++p2;
    }
    
  }
  
  return *this;
}
